; Author: Juergen Lorenz ; ju (at) jugilo (dot) de
;
; Copyright (c) 2013-2015, Juergen Lorenz
; All rights reserved.
; 
; Redistribution and use in source and binary forms, with or without
; modification, are permitted provided that the following conditions are
; met:
; 
; Redistributions of source code must retain the above copyright
; notice, this list of conditions and the following dispasser.
; 
; Redistributions in binary form must reproduce the above copyright
; notice, this list of conditions and the following dispasser in the
; documentation and/or other materials provided with the distribution.
; 
; Neither the name of the author nor the names of its contributors may be
; used to endorse or promote products derived from this software without
; specific prior written permission. 
; 
; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
; IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
; TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
; PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
; HOLDERS OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
; SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
; TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
; PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
; LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
; NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
; SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

#|[
The fundamental binding-construct, bind, is patterned after Paul Graham's
dbind, cf. "On Lisp", p. 232.
In Chicken, dbind for lists could look as follows

  (define-syntax dbind
    (ir-macro-transformer
      (lambda (form inject compare?)
        (letrec (
          (mappend
            (lambda (fn lists)
              (apply append (map fn lists))))
          (destruc
            (lambda (pat seq)
              (let loop ((pat pat) (seq seq) (n 0))
                (if (pair? pat)
                  (let ((p (car pat)) (recu (loop (cdr pat) seq (+ n 1))))
                    (if (symbol? p)
                      (cons `(,p (list-ref ,seq ,n)) recu)
                      (let ((g (gensym)))
                        (cons (cons `(,g (list-ref ,seq ,n))
                                    (loop p g 0))
                              recu))))
                  (if (null? pat)
                    '()
                    `((,pat (list-tail ,seq ,n))))))))
          (dbind-ex
            (lambda (binds body)
              (if (null? binds)
                `(begin ,@body)
                `(let ,(map (lambda (b) (if (pair? (car b)) (car b) b))
                            binds)
                   ,(dbind-ex
                      (mappend (lambda (b) (if (pair? (car b)) (cdr b) '()))
                               binds)
                      body)))))
          )         
          (let ((pat (cadr form))
                (seq (caddr form))
                (body (cdddr form))
                (gseq 'seq))
            `(let ((,gseq ,seq))
               ,(dbind-ex (destruc pat gseq) body)))))))

This code works as follows: First, destruc traverses the pattern and
groups each symbol with some list accessing code, using gensyms to step
down the pattern while grouping the gensym bound object with all pairs
depending on this gensym. So, for example,

  (destruc '(a (b . c) . d) 'seq)

will result in

  ((a (list-ref seq 0))
   ((#:g (list-ref seq 1)) (b (list-ref #:g 0)) (c (list-tail #:g 1)))
   (d (list-tail seq 2)))

This tree is then transformed via dbind-ex into a nested let

  (let ((a (list-ref seq 0))
        (#:g (list-ref seq 1))
        (d (list-tail seq 2)))
    (let ((b (list-ref #:g 0))
          (c (list-tail #:g 1)))
      body))
 
Note, that the destructuring procedures are local to this macro. This is
necessary in Chicken for the macro to work, in particular in compiled
code, unless you import them for-syntax. But since they are of no
interest outside of the macro, local procedrues are preferable.

Note further, that ir-macro-transformer does all the necessary renaming
transparently behind the scene, even if the helpers where defined in
another module. In particular, gseq needn't be a gensym.

And note, that Graham's code didn't check for seq's length, i.e. 
(dbind (a b) '(1 2 3) (list a b) would happily return '(1 2).

Graham's original code works on the sequence datatype, so vectors and
strings are destructured as well. Sequences don't exist in Scheme,
unless you import-for-syntax Felix' sequences egg. To make this module
self-contained, I prefer to supply access-routines closed over a table,
which provides sequence versions of list-ref and list-tail, the only
sequence routines used by destruc above, as well as a sequence version
of length, which is needed to do the length checks.

There are some features, which I would like to have and which are
implemented as well. First wildcards, represented by the underscore
symbol. It matches everything, but binds nothing. So it can appear
multiple times in the same macro. Wildcard symbols are simply not
collected in the destruc routine.

Second, non-symbol literals, which don't bind anything, of course, but
match only themselves. This and the length checks are treated simply by
pairing them as well with check-routines in destruc but separating the
pairs with leading symbol from those with leading nil or literal in
dbind-ex. The former are bound with lets as in Graham's code, the
latter's cadrs being evaluated before the recursive call to dbind-ex.

The last feature missing is fenders, which is important in particular
for bind-case and can easily be implemented with a where clause: A
pattern matches successfully if only each pattern variable can be bound
and the where clause is satisfied. If the where clause doesn't pass, the
next pattern is tried in bind-case or a bind-exception is signalled in
bind.

bind-case is the macro, which is heavily used in macro-rules. It does
all the destructuring there while implicit renaming cares for variable
captures.

bind and bind-case are implemented in the first module, on which the
others rely.

Note, that the implementation of define-macro and macro-rules in the
macro-bindings module is surprisingly easy having implicit-renaming
macros and binding routines at ones disposal.
]|#

(module basic-bindings
  (bind bind-case
   bind-seq-length bind-seq-ref bind-seq-tail bind-table-show bind-table-add!
   bind-exception bind-exception-handler signal-bind-exception
   list-of vector-of symbol-dispatcher basic-bindings)
  (import scheme
          (only chicken case-lambda condition-case define-values
                error subvector
                current-exception-handler condition-predicate
                get-condition-property make-property-condition
                make-composite-condition signal abort print)
          (only data-structures conjoin list-of?))
  (import-for-syntax
    (only data-structures compress))

#|[
Let's start with defining bind-exceptions, a corresponding exception-handler,
and registering this handler
]|#

;;; (bind-exception loc msg arg ...)
;;; --------------------------------
;;; composite condition, to allow for (exn bind) in condition-case
(define (bind-exception loc msg . args)
  (make-composite-condition
    (make-property-condition 'exn
      'location `(,loc)
      'message msg
      'arguments (apply list args))
    (make-property-condition 'bind
      'location `(,loc)
      'message msg
      'arguments (apply list args))))

;;; (signal-bind-exception loc msg arg ...)
;;; ---------------------------------------
;;; signals a bind-exception, can be used instead of error
(define (signal-bind-exception loc msg . args)
  (signal
    (apply bind-exception loc msg args)))

;;; (bind-exception-handler var)
;;; ----------------------------
;;; exception-handler to be passed to the parameter
;;; current-exception-handler
(define bind-exception-handler
  (let ((old-handler (current-exception-handler)))
    (lambda (var)
      (if ((condition-predicate 'bind) var)
        (begin
          (display "Bind error: ")
          (print (get-condition-property var 'bind 'location))
          (print (get-condition-property var 'bind 'message))
          (for-each print (get-condition-property var 'bind 'arguments))
          (abort (make-property-condition 'exn
                   'message "exception-handler returned")))
        (old-handler var)))))

;;; set current-exception-handler
(current-exception-handler bind-exception-handler)

#|[ 
The following three routines maintain the lookup table for the needed
sequence primitives. Instead of bind-table-lookup the three sequence
primitives below are exported.
If you prefer, you can use the sequence primitives size, elt and sub of
the sequences egg, provided you rename them bind-seq-length,
bind-seq-ref and bind-seq-tail respectively.
]|#

;;; (bind-table-lookup obj)
;;; -----------------------
;;; returns an association list of predicates and associated vectors
;;; with length, ref and tail primitives
;;;
;;; (bind-table-show)
;;; -----------------
;;; prints the contents of the table
;;;
;;; (bind-table-add! type? len ref tail)
;;; ------------------------------------
;;; adds a new list to the top of the table
(define-values (bind-table-lookup bind-table-show bind-table-add!)
  (let (
    (tbl
      (list (cons pair?
                  (vector
                    (lambda (obj)
                      (let loop ((obj obj) (len 0))
                        (if (pair? obj)
                          (loop (cdr obj) (+ len 1))
                          len)))
                    list-ref
                    list-tail))
            (cons vector?
                  (vector vector-length vector-ref subvector))
            (cons string?
                  (vector string-length string-ref substring))
            ;; atoms catch all
            (cons (lambda (obj) (not (pair? obj)))
                  (vector
                    (lambda (obj) 0)
                    (lambda (obj pos)
                      (signal-bind-exception 'bind-table-lookup
                                             "out of range"
                                             obj
                                             pos))
                    (lambda (obj pos)
                      (if (zero? pos)
                        obj
                        (signal-bind-exception 'bind-table-lookup
                                               "out of range"
                                               obj)))))
            ))
    )
    (values
      (lambda (obj)
        (let loop ((tbl tbl))
          ;; note, that we have a catch-all predicate in the table
          (if ((caar tbl) obj)
            (cdar tbl)
            (loop (cdr tbl)))))
      (lambda () (print tbl))
      (lambda (type? len ref tail)
        (set! tbl (cons (cons type?
                              (vector len ref tail))
                        tbl))))
      ))

;;; (bind-seq-length seq)
;;; ---------------------
;;; returns the length of the sequence argument
(define (bind-seq-length obj)
  ((vector-ref (bind-table-lookup obj) 0) obj))

;;; (bind-seq-ref seq pos)
;;; ----------------------
;;; returns the item of the sequence argument at index pos
(define (bind-seq-ref obj pos)
  ((vector-ref (bind-table-lookup obj) 1) obj pos))

;;; (bind-seq-tail seq pos)
;;; -----------------------
;;; returns the tail of the sequence argument starting at index pos
(define (bind-seq-tail obj pos)
  ((vector-ref (bind-table-lookup obj) 2) obj pos))

#|[
The following is Graham's dbind extended with fenders, wildcards,
non-symbol literals and length-checks. For example

  (bind (x (y z)) '(1 #(2 3)) (where (integer? x)) (list x y z))

will result in '(1 2 3) while

  (bind (_ ("y" z)) '(1 #("y" z)) z)

will produce 3

]|#

;;; (bind pat seq (where . fenders) .. xpr ....)
;;; ---------------------------------------------
;;; binds pattern variables of pat to corresponding subexpressions of
;;; seq and executes body xpr .... in this context, provided all
;;; fenders pass
(define-syntax bind
  (ir-macro-transformer
    (lambda (form inject compare?)
      (letrec (
        (len 'bind-seq-length)
        (ref 'bind-seq-ref)
        (tail 'bind-seq-tail)
        (filter2
          (lambda (ok? lst)
            (let loop ((lst lst) (yes '()) (no '()))
              (if (null? lst)
                (list (reverse yes) (reverse no))
                (let ((first (car lst)) (rest (cdr lst)))
                  (if (ok? first)
                    (loop rest (cons first yes) no)
                    (loop rest yes (cons first no))))))))
        (mappend
          (lambda (fn lists)
            (apply append (map fn lists))))
        (destruc
          (lambda (pat seq)
            (let loop ((pat pat) (seq seq) (n 0))
              (cond
                ((pair? pat)
                 (let ((p (car pat)) (recu (loop (cdr pat) seq (+ n 1))))
                   (cond
                     ((symbol? p)
                      (if (compare? p '_)
                        ;; skip
                        recu
                        (cons `(,p (,ref ,seq ,n)) recu)))
                     ((pair? p)
                      (let ((g (gensym)))
                        (cons (cons `(,g (,ref ,seq ,n))
                                    (loop p g 0))
                              recu)))
                     (else
                       (cons `(,p (equal? ',p (,ref ,seq ,n)))
                             recu))
                     )))
                ((symbol? pat)
                 `((,pat (,tail ,seq ,n))))
                ((null? pat)
                 `((,pat (zero? (,len (,tail ,seq ,n))))))
                ))))
        (dbind-ex
          (lambda (binds body)
            (if (null? binds)
              `(begin ,@body)
              (apply (lambda (defs checks)
                       `(let ,defs
                          (if (and ,@(map cadr checks))
                            ,(dbind-ex
                               (mappend (lambda (b) (if (pair? (car b)) (cdr b) '()))
                                        binds)
                               body)
                            (signal-bind-exception
                              'bind
                              "match error"
                              ',(cons 'and (map cadr checks))))
                          ))
                     (filter2 (lambda (pair) (symbol? (car pair)))
                              (map (lambda (b) (if (pair? (car b)) (car b) b))
                                   binds)))
              )))
        )         
        (let ((pat (cadr form))
              (seq (caddr form))
              (xpr (cadddr form))
              (xprs (cddddr form))
              (gseq 'seq))
          (let ((fender? (and (pair? xpr)
                              (compare? 'where (car xpr))))
                (destruc-pat-gseq (destruc pat gseq)))
            (if fender?
              `(let ((,gseq ,seq))
                 (if ,(dbind-ex destruc-pat-gseq
                                (list (cons 'and (cdr xpr))))
                   ,(dbind-ex destruc-pat-gseq xprs)
                   (signal-bind-exception 'bind
                                          "match error"
                                          ,gseq
                                          ',pat
                                          ',xpr)))
              `(let ((,gseq ,seq))
                 ,(dbind-ex destruc-pat-gseq (cons xpr xprs))))
            ))))))

#|[
The following macro does more or less the same what the match macro from
the matchable package does, for example

  (bind-case '(1 (2 3))
    ((x y) (where (list? y)) (list x y))
    ((x (y . z)) (list x y z))
    ((x (y z)) (list x y z))) ;-> '(1 2 (3))

or, to give a more realistic example, mapping:

  (define (my-map fn lst)
    (bind-case lst
      (() '())
      ((x . xs) (cons (fn x) (my-map fn xs)))))

To improve error messages, we wrap it around an inner version,
bind-case-inner, which does all of the work.
]|#

;;; inner version, not exported
(define-syntax bind-case-inner
  (ir-macro-transformer
    (lambda (form inject compare?)
      (let ((seq (cadr form)) (clauses (cddr form)))
        (if (null? clauses)
          `(signal-bind-exception 'bind-case-inner
                                  "no match for"
                                  ,seq
                                  )

          `(condition-case (bind ,(caar clauses) ,seq ,@(cdar clauses))
             ((exn type)
              (bind-case ,seq ,@(cdr clauses)))
             ((exn bind)
              (bind-case ,seq ,@(cdr clauses)))))))))

;;; (bind-case seq (pat (where fender ...) .. xpr ....) ....)
;;; ---------------------------------------------------------
;;; Checks if seq matches pattern pat [satisfying fender ...] ....
;;; in sequence, binds the pattern variables of the first matching
;;; pattern to corresponding subexpressions of seq and executes
;;; corresponding body xpr ....
(define-syntax bind-case
  (ir-macro-transformer
    (lambda (form inject compare?)
      (let ((seq (cadr form)) (clauses (cddr form)))
        `(condition-case (bind-case-inner ,seq ,@clauses)
           ((exn bind)
            (signal-bind-exception 'bind-case
                                   "no match for"
                                   ,seq
                                   ',(cons 'in
                                           (map (lambda (clause)
                                                  (list (car clause)
                                                        (if (and (pair? (cadr clause))
                                                                 (compare? (caadr clause) 'where))
                                                          (cadr clause)
                                                          (list 'where #t))))
                                          clauses)))))))))

#|[
At last some helper functions, which sometimes make life easier
]|#

;;; (list-of ok? ....)
;;; ------------------
;;; returns a list predicate which checks all ok? arguments
(define (list-of . oks?) (list-of? (apply conjoin oks?)))

;;; (vector-of ok? ....)
;;; --------------------
;;; returns a list predicate which checks all ok? arguments
(define (vector-of . oks?)
  (let (
    (vector-of?
      (lambda (ok?)
        (lambda (vec)
          (and (vector? vec)
               (let loop ((n 0))
                 (cond
                   ((= n (vector-length vec))
                    #t)
                   ((ok? (vector-ref vec n))
                    (loop (+ n 1)))
                   (else #f)))))))
    )
    (vector-of? (apply conjoin oks?))))

;;; (symbol-dispatcher alist)
;;; -------------------------
;;; returns a procedure of zero or one argument, which shows all cars
;;; or the cdr of the alist item with car symbol
(define (symbol-dispatcher alist)
  (case-lambda
    (()
     (map car alist))
    ((sym)
     (let ((pair (assq sym alist)))
       (if pair
         (for-each print (cdr pair))
         (error "Not in list"
                sym
                (map car alist)))))))

;;; (basic-bindings sym ..)
;;; -----------------------
;;; documentation procedure of this module
(define basic-bindings
  (symbol-dispatcher '(
    (bind
      macro:
      (bind pat seq (where . fenders) .. xpr ....)
      "a variant of Common Lisp's destructuring-bind")
    (bind-case
      macro:
      (bind-case seq (pat (where . fenders) .. xpr ....) ....)
      "matches seq against pat with optional fenders in a case regime")
    (bind-exception
      procedure:
      (bind-exception loc msg arg ...)
      "generates a composite condition with location symbol, string message"
      "and passible additional arguments arg ...")
    (signal-bind-exception
      procedure:
      (bind-exception loc msg arg ...)
      "signals a composite condition with location symbol, string message"
      "and passible additional arguments arg ...")
    (bind-exception-handler
      procedure:
      (bind-exception-handler var)
      "to be passed to the parameter current-exception-handler")
    (bind-seq-length
      procedure:
      (bind-seq-length seq)
      "redurns the length of a sequence")
    (bind-seq-ref
      procedure:
      (bind-seq-ref seq pos)
      "returns the item at position pos of a sequence")
    (bind-seq-tail
      procedure:
      (bind-seq-ref seq pos)
      "returns the tail starting at position pos of a sequence")
    (bind-table-show
      procedure:
      (bind-table-show)
      "pretty prints the sequence table")
    (bind-table-add!
      procedure:
      (bind-table-add! type? len ref tail)
      "adds a new table item to the front of the sequence table")
    (list-of
      procedure:
      (list-of ok? ...)
      "generates a list predicate which checks all of its arguments")
    (vector-of
      procedure:
      (vector-of ok? ...)
      "generates a list predicate which checks all of its arguments")
    (symbol-dispatcher
      procedure:
      (symbol-dispatcher alist)
      "generates a procedure of zero or one argument showing all"
      "cars or the cdr or the alist item with symbol as car")
    )))

) ; basic-bindings


#|[
Now we'll use bind-case to create procedural macros.  The first,
macro-rules, is a procedural version of syntax-rules. It is as
convenient as the latter, but much more powerfull. For example, it can
use injected symbols and once pattern variables, do some of its work at
compile-time, use local functions at compile-time and what have you.
Contrary to syntax-rules the templates usually evaluate to quasiquoted
expressions.
Other procedural macro-building routines are provided as well, in
particular, a hygienic define-macro, based on bind as well as local
versions of it, macro-let and macro-letrec.
]|#

(module macro-bindings
  (define-macro macro-rules macro-let macro-letrec once-only
    ;mac-rules
   with-gensyms macro-bindings)
  (import scheme basic-bindings)
  (import-for-syntax
    (only basic-bindings bind bind-case)
    (only data-structures compress))


;;; (macro-rules sym ... (key ...) (pat (where fender ...) .. tpl) ....)
;;; --------------------------------------------------------------------
;;; where sym ... are injected non-hygienig symbols, key ... are
;;; additional keywords, pat ....  are nested lambda-lists without
;;; spezial meaning of ellipses and tpl .... usually evaluate to
;;; quasiquoted templates. The optional fenders belong to the pattern
;;; matching process.
(define-syntax macro-rules
  (ir-macro-transformer
    (lambda (f i c?)
      (let ((f* (let loop ((tail (cdr f)) (head '()))
                  (if (symbol? (car tail))
                    (loop (cdr tail) (cons (car tail) head))
                    (cons head tail)))))
        (let ((syms (car f*))
              (keys (cadr f*))
              (rules (cddr f*))
              (flatten*
                ; imported flatten doesn't work with pseudo-lists
                (lambda (tree)
                  (let loop ((tree tree) (result '()))
                    (cond
                      ((pair? tree)
                       (loop (car tree) (loop (cdr tree) result)))
                      ((null? tree) result)
                      (else
                        (cons tree result)))))))
          `(ir-macro-transformer
             (lambda (form inject compare?)
               (let ,(map (lambda (s)
                            `(,s (inject ',s)))
                          syms)
                 (bind-case form
                   ,@(map (lambda (rule)
                            (let* ((pat (car rule))
                                   (fpat (flatten* pat))
                                   (kpat (compress
                                           (map (lambda (x) 
                                                  (memq x keys))
                                                fpat)
                                           fpat))
                                   ;; compare? keywords with its names
                                   (key-checks
                                     (map (lambda (p s)
                                            `(compare? ,p ,s))
                                          kpat
                                          (map (lambda (x) `',x)
                                               kpat))))
                              (let ((tpl (cdr rule)))
                                ;; add key-checks to where clause of tpl
                                (if (and (pair? (car tpl)) 
                                         (c? (caar tpl) 'where))
                                  `(,pat (where ,@key-checks ,@(cdar tpl))
                                         ,@(cdr tpl))
                                  `(,pat (where ,@key-checks) ,@tpl)))))
                          rules))))))))))

;;; (define-macro (name . args) (where fender ...) .. xpr ....)
;;; -----------------------------------------------------------
;;; simple hygienic macro without injections and keywords, but with
;;; fenders and once arguments.
(define-syntax define-macro
  (ir-macro-transformer
    (lambda (f i c?)
      (let ((code (cadr f))
            (xpr (caddr f))
            (xprs (cdddr f)))
        `(define-syntax ,(car code)
           (ir-macro-transformer
             (lambda (form inject compare?)
               (bind ,(cdr code) (cdr form) ,xpr ,@xprs))))))))

#|[
Now follow the local versions of define-macro, macro-let and
macro-letrec. Since the syntax of both is identical, they are
implemented by means of a helper macro.
]|#

;; helper for macro-let and macro-letrec
(define-syntax macro
  (ir-macro-transformer
    (lambda (form inject compare?)
      (let ((op (cadr form))
            (pat-tpl-pairs (caddr form))
            (xpr (cadddr form))
            (xprs (cddddr form)))
        (let ((pats (map car pat-tpl-pairs))
              (tpls (map cdr pat-tpl-pairs)))
          `(,op ,(map (lambda (pat tpl)
                               `(,(car pat)
                                  (macro-rules ()
                                     ((_ ,@(cdr pat)) ,@tpl))))
                              pats tpls)
                       ,xpr ,@xprs))))))

;;; (macro-let (((name . args) (where fender ...) .. xpr ...) ...) body ....)
;;; -------------------------------------------------------------------------
;;; evaluates body ... in the context of parallel macros name ....
(define-syntax macro-let
  (ir-macro-transformer
    (lambda (form inject compare?)
      (let ((pat-tpl-pairs (cadr form))
            (xpr (caddr form))
            (xprs (cdddr form)))
        `(macro let-syntax ,pat-tpl-pairs ,xpr ,@xprs)))))

;;; (macro-letrec (((name . args) (where fender ...) .. xpr ...) ...) body ....)
;;; ----------------------------------------------------------------------------
;;; evaluates body ... in the context of recursive macros name ....
(define-syntax macro-letrec
  (ir-macro-transformer
    (lambda (form inject compare?)
      (let ((pat-tpl-pairs (cadr form))
            (xpr (caddr form))
            (xprs (cdddr form)))
        `(macro letrec-syntax ,pat-tpl-pairs ,xpr ,@xprs)))))

;;; (with-gensyms (name ....) xpr ....)
;;; -----------------------------------
;;; binds name ... to (gensym 'name) ... in body xpr ...
(define-syntax with-gensyms
  (ir-macro-transformer
    (lambda (form inject compare?)
      `(let ,(map (lambda (n) `(,n (gensym ',n))) (cadr form))
         ,@(cddr form)))))

;;; (once-only (x ....) xpr ....)
;;; -----------------------------
;;; macro-arguments x .... are only evaluated once and from left to
;;; right in the body xpr ....
;;; The code is more or less due to
;;; P. Seibel, Practical Common Lisp, p. 102
(define-syntax once-only
  (ir-macro-transformer
    (lambda (form inject compare?)
      (let ((names (cadr form))
            (body (cddr form)))
        (let ((gensyms (map (lambda (x) (gensym)) names)))
          `(let ,(map (lambda (g) `(,g (gensym))) gensyms)
             `(let ,(list ,@(map (lambda (g n) ``(,,g ,,n))
                                 gensyms names))
                ,(let ,(map (lambda (n g) `(,n ,g))
                            names gensyms)
                   ,@body))))))))

;;; (macro-bindings sym ..)
;;; -----------------------
;;; documentation procedure.
(define macro-bindings
  (symbol-dispatcher '(
    (macro-rules
      macro:
      (macro-rules literal ... (keyword ...) (pat tpl) ....)
      "procedural version of syntax-rules"
      "with optional injected literals"
      "and quasiquoted templates")
    (define-macro
      macro:
      (define-macro (name . args) xpr ....)
      "a version of macro-rules with only one rule"
      "no injected symbols and no keywords")
    (macro-let
      macro:
      (macro-let (((name . args) (where fender ...) .. xpr ...) ...) body ....)
      "evaluates body ... in the context of parallel macros name ....")
    (macro-letrec
      macro:
      (macro-letrec (((name . args) (where fender ...) .. xpr ...) ...) body ....)
      "evaluates body ... in the context of recursive macros name ....")
    (once-only
      macro:
      (once-only (x ....) xpr ....)
      "arguments x ... are evaluated only once and"
      "from left to right in the body xpr ....")
    (with-gensyms
      macro:
      (with-gensyms (x ....) xpr ....)
      "generates a series of gensyms x .... to be used in body xpr ...")
    )))
) ; macro-bindings

#|[
The binding macros to follow are all implemented either with
define-macro or with macro-rules, hence the latter must be
imported for-syntax. Had we chosen the direct implementation of
define-macro with bind instead of macro-rules, we had to import
for-syntax bind as well. This is the reason, why separate modules are
needed. Note, that the fundamental macros, bind and bind-case, know how
to handle where clauses. Hence the derived macros do as well.
]|#

(module more-bindings *
  (import scheme
          basic-bindings
          (except macro-bindings macro-rules)
          )
  (import-for-syntax (only basic-bindings bind)
                     (only macro-bindings macro-rules))

#|[
The next macro, bindable?, can be used to check, if a
sequence-expression matches a pattern and passes all fenders.
]|#

;;; (bindable? pat (where fender ...) ..)
;;; -------------------------------------
;;; returns a unary predicate which checks, if its argument matches pat
;;; and fulfills the predicates in the list fender ...
;;; Mostly used in fenders of macro-rules and define-macro, but must
;;; then be imported for-syntax.
(define-syntax bindable?
  (macro-rules (where)
    ((_ pat (where . fenders))
     `(lambda (seq)
        (condition-case (bind ,pat seq (and ,@fenders))
          ((exn bind) #f))))
    ((_ pat)
     `(bindable? ,pat (where)))))

#|[ 
The following two macros, bind-define and bind-set!, destructure their
sequence arguments with respect to their pattern argument and define or
set! the pattern variables correspondingly.  For example, one can define
multiple procedures operating on a common state

  (bind-define (push top pop)
    (let ((state '()))
      (list
        (lambda (arg) (set! state (cons arg state)))
        (lambda () (car state))
        (lambda () (set! state (cdr state))))))

]|#

;; helper macro for bind-define and bind-set!
(define-macro (bind-def-set! pat seq def?)
  (let ((sym? (lambda (p)
                (and (symbol? p)
                     (not (compare? p '_))))))
    (let ((aux (let copy ((pat pat))
                 (cond
                   ((sym? pat) (gensym))
                   ((pair? pat)
                    (cons (copy (car pat)) (copy (cdr pat))))
                   (else pat))))
          (flatten*
            ; imported flatten doesn't work with pseudo-lists
            (lambda (tree)
              (let loop ((tree tree) (result '()))
                (cond
                  ((pair? tree)
                   (loop (car tree) (loop (cdr tree) result)))
                  ((null? tree) result)
                  (else
                    (cons tree result))))))
          (filter
            (lambda (ok? lst)
              (compress (map ok? lst) lst))))
      (if def?
        `(begin
           ,@(map (lambda (p) `(define ,p ',p))
                  (filter sym? (flatten* pat)))
           (bind ,aux ,seq
             ,@(map (lambda (p a) `(set! ,p ,a))
                    (filter sym? (flatten* pat))
                    (filter sym? (flatten* aux)))))
        `(begin
           (bind ,aux ,seq
             ,@(map (lambda (p a) `(set! ,p ,a))
                    (filter sym? (flatten* pat))
                    (filter sym? (flatten* aux))))))
      )))


;;; (bind-define pat seq)
;;; ---------------------
;;; destructures the sequence seq according to the pattern pat and sets
;;; pattern variables with values corresponding to subexpressions of seq
(define-macro (bind-define pat seq)
  `(bind-def-set! ,pat ,seq #t))

;;; (bind-set! pat seq)
;;; -------------------
;;; sets pattern variables of pat to corresponding sub-expressins of seq
(define-macro (bind-set! pat seq)
  `(bind-def-set! ,pat ,seq #f))

#|[
Now we can define two macros, which simply combine lambda with
bind, the first destructures simply one argument, the second a
whole list. An example of a call and its result is

  ((bind-lambda (a (b . c) . d) (list a b c d))
   '(1 #(20 30 40) 2 3))
  -> '(1 20 #(30 40) (2 3)))))

  ((bind-lambda* ((a (b . c) . d) (e . f))
     (list a b c d e f))
   '(1 #(20 30 40) 2 3) '#(4 5 6))
  -> '(1 20 #(30 40) (2 3) 4 #(5 6)))
]|#

;;; (bind-lambda pat (where fender ...) .. xpr ....)
;;; ------------------------------------------------
;;; combination of lambda and bind, one pattern argument
(define-macro (bind-lambda pat xpr . xprs)
  `(lambda (x) (bind ,pat x ,xpr ,@xprs)))

;;; (bind-lambda* pat (where fender ...) .. xpr ....)
;;; -------------------------------------------------
;;; combination of lambda and bind, multiple pattern arguments
(define-macro (bind-lambda* pat xpr . xprs)
  `(lambda x (bind ,pat x ,xpr ,@xprs)))

#|[
The following macro, bind*, is a named version of bind. It takes an
additional argument besides those of bind, which is bound to a
recursive procedure, which can be called in bind's body. The pattern
variables are initialised with the corresponding subexpressions in seq.
For example

  (bind* loop (x y) '(5 0)
    (if (zero? x)
      (list x y)
      (loop (list (sub1 x) (add1 y)))))
  -> '(0 5)
]|#

;;; (bind* name pat seq (where . fenders) .. xpr ....)
;;; ---------------------------------------------------
;;; named version of bind
(define-macro (bind* name pat seq xpr . xprs)
  `((letrec ((,name 
               (bind-lambda ,pat ,xpr ,@xprs)))
      ,name)
    ,seq))

#|[
Now the implementation of a nested version of let, named and unnamed,
is easy: Simply combine bind and bind*. For example

  (bind-let (
     (((x y) z) '((1 2) 3))
     (u (+ 2 2))
     ((v w) '(5 6))
     )
     (list x y z u v w))
  -> '(1 2 3 4 5 6)

  (bind-let loop (((a b) '(5 0)))
    (if (zero? a)
      (list a b)
      (loop (list (sub1 a) (add1 b)))))
  -> '(0 5)
]|#

;;; (bind-let loop .. ((pat seq) ...) xpr ....)
;;; --------------------------------------------
;;; nested version of let, named and unnamed
(define-syntax bind-let
  (macro-rules ()
    ((_ loop () xpr . xprs)
     `(let ,loop () ,xpr ,@xprs))
    ((_ loop ((pat0 seq0) . pat-seq-pairs) xpr . xprs)
     `(bind* ,loop
        ,(cons pat0 (map car pat-seq-pairs))
        (list ,seq0 ,@(map cadr pat-seq-pairs))
        ,xpr ,@xprs))
    ((_ () xpr . xprs)
     `(begin ,xpr ,@xprs))
    ((_ ((pat0 seq0) . pat-seq-pairs) xpr . xprs)
     `(bind
        ,(cons pat0 (map car pat-seq-pairs))
        (list ,seq0 ,@(map cadr pat-seq-pairs))
        ,xpr ,@xprs))
    ))

#|[
The sequential version of bind-let should work as follows

  (bind-let* (
     (((x y) z) '((1 2) 3))
     (u (+ 1 2 x))
     ((v w) (list (+ z 2) 6))
     )
     (list x y z u v w))
  -> '(1 2 3 4 5 6)
]|#

;;; (bind-let* ((pat seq) ...) xpr ....)
;;; -------------------------------------
;;; sequential version of bind-let
(define-syntax bind-let*
  (macro-rules ()
    ((_ () xpr . xprs)
     `(let () ,xpr ,@xprs))
    ((_ ((pat seq) . pat-seq-pairs) xpr . xprs)
     `(bind ,pat ,seq (bind-let* ,pat-seq-pairs ,xpr ,@xprs)))))

#|[
And here is the recursive version of bind.

  (bindrec ((o?) e?)
    (list (list (lambda (m) (if (zero? m) #f (e? (- m 1)))))
          (lambda (n) (if (zero? n) #t (o? (- n 1)))))
    (list (o? 95) (e? 95)))
  -> '(#t #f)
]|#

;;; (bindrec pat seq (where . fenders) .. xpr ....)
;;; -----------------------------------------------
;;; recursive version of bind
(define-macro (bindrec pat seq xpr . xprs)
  `(bind ,pat ',pat
     ; bind pattern variables to auxiliary values
     ; so that they are in scope
     (bind-set! ,pat ,seq)
     ; set! the real values
     ,xpr ,@xprs))

#|[
The recursive version of bind-let works as follows
  
  (bind-letrec (
    ((o? (e?))
     (list (lambda (m) (if (zero? m) #f (e? (- m 1))))
           (vector (lambda (n) (if (zero? n) #t (o? (- n 1)))))))
    )
    (list (o? 95) (e? 95)))
  -> '(#t #f)
]|#

;;; (bind-letrec ((pat seq) ...) xpr ....)
;;; ---------------------------------------
;;; recursive version of bind-let
(define-macro (bind-letrec pat-seq-pairs xpr . xprs)
  `(bindrec ,(map car pat-seq-pairs)
     (list ,@(map cadr pat-seq-pairs))
     ,xpr ,@xprs))

#|[ 
The next two macros combine lambda and bind-case and do more or less the
same as match-lambda and match-lambda* in the matchable package. The
first destructures one argument, the second a list of arguments.
Here is an example together with its result:

  ((bind-case-lambda
     ((a (b . c) . d) (list a b c d))
     ((e . f) (where (zero? e)) e)
     ((e . f) (list e f)))
   '(1 2 3 4 5))
  -> '(1 (2 3 4 5))

  ((bind-case-lambda*
     (((a (b . c) . d) (e . f))
      (list a b c d e f)))
   '(1 #(20 30 40) 2 3) '(4 5 6))
  -> '(1 20 #(30 40) (2 3) 4 (5 6))
]|#

;;; (bind-case-lambda (pat (where . fenders) .. xpr ....) ....)
;;; ------------------------------------------------------------
;;; combination of lambda and bind-case, one pattern argument
(define-syntax bind-case-lambda
  (macro-rules ()
    ((_ (pat xpr . xprs))
     `(lambda (x) (bind ,pat x ,xpr ,@xprs)))
    ((_ clause . clauses)
     `(lambda (x)
       (bind-case x ,clause ,@clauses)))))

;;; (bind-case-lambda* (pat (where . fenders) .. xpr ....) ....)
;;; -------------------------------------------------------------
;;; combination of lambda and bind-case, multiple pattern arguments
(define-syntax bind-case-lambda*
  (macro-rules ()
    ((_ (pat xpr . xprs))
     `(lambda x (bind ,pat x ,xpr ,@xprs)))
    ((_ clause . clauses)
     `(lambda x
       (bind-case x ,clause ,@clauses)))))

#|[ 
The following macro is sometimes named let/cc or let-cc
]|#

;;; (bind/cc cc xpr ....)
;;; ---------------------
;;; captures the current continuation, binds it to cc and executes
;;; xpr .... in this context
(define-macro (bind/cc cc xpr . xprs)
  `(call-with-current-continuation
     (lambda (,cc) ,xpr ,@xprs)))

;;; (more-bindings sym ..)
;;; ----------------------
;;; documentation procedure
(define more-bindings
  (symbol-dispatcher '(
    (bindable?
      macro:
      (bindable? pat (where . fenders) ..)
      "returns a unary predicate, which checks"
      "if its argument matches pat and passes all fenders")
    (bind-set!
      macro:
      (bind-set! pat seq)
      "sets multiple variables by destructuring its sequence argument")
    (bind-define
      macro:
      (bind-define pat seq)
      "defines multiple variables by destructuring its sequence argument")
    (bind-lambda
      macro:
      (bind-lambda pat (where . fenders) .. xpr ....)
      "combination of lambda and bind, one pattern argument")
    (bind-lambda*
      macro:
      (bind-lambda* pat (where . fenders) .. xpr ....)
      "combination of lambda and bind, multiple pattern arguments")
    (bind*
      macro:
      (bind* loop pat seq (where . fenders) .. xpr ....)
      "named version of bind")
    (bind-let
      macro:
      (bind-let loop .. ((pat seq) ...) xpr ....)
      "nested version of let, named and unnamed")
    (bind-let*
      macro:
      (bind-let* ((pat seq) ...) xpr ....)
      "nested version of let*")
    (bindrec
      macro:
      (bindrec pat seq (where . fenders) .. xpr ....)
      "recursive version of bind")
    (bind-letrec
      macro:
      (bind-letrec ((pat seq) ...) xpr ....)
      "recursive version of bind-let")
    (bind-case-lambda
      macro:
      (bind-case-lambda (pat (where . fenders) .. xpr ....) ....)
      "combination of lambda and bind-case with one pattern argument")
    (bind-case-lambda*
      macro:
      (bind-case-lambda* (pat (where . fenders) .. xpr ....) ....)
      "combination of lambda and bind-case with multiple pattern arguments")
    (bind/cc
      macro:
      (bind/cc cc xpr ....)
      "binds cc to the current contiunation"
      "and execute xpr ... in this context")
    )))

) ; more-bindings

#|[ 
And now we put all three modules into one for convenience
]|#

(module bindings *
  (import scheme
          (only chicken case-lambda error)
          basic-bindings macro-bindings more-bindings)
  (reexport basic-bindings macro-bindings more-bindings)

;;; (bindings sym ..)
;;; -----------------
;;; documentation procedure.
(define bindings
  (let ((lst (append (basic-bindings)
                     (macro-bindings)
                     (more-bindings))))
    (case-lambda
      (() lst)
      ((sym)
       (cond
         ((memq sym (basic-bindings))
          (basic-bindings sym))
         ((memq sym (macro-bindings))
          (macro-bindings sym))
         ((memq sym (more-bindings))
          (more-bindings sym))
         (else
           (error "Not in list" sym lst)))))))
) ; bindings

